@@ -1,5 +1,5 @@
 /*******************************************************************************
- * Copyright (c) 2017, College of William & Mary
+ * Copyright (c) 2018, College of William & Mary
  * All rights reserved.
  * 
  * Redistribution and use in source and binary forms, with or without
@@ -68,13 +68,6 @@ typedef enum {         /* Initially fill up the search subspace with: */
    primme_init_user    /* c) provided vectors or a single random vector */
 } primme_init;
 
-
-typedef enum {
-   primme_thick,
-   primme_dtr
-} primme_restartscheme;
-
-
 typedef enum {
    primme_full_LTolerance,
    primme_decreasing_LTolerance,
@@ -90,7 +83,9 @@ typedef enum {
    primme_event_restart,            /* report at every basis restart          */
    primme_event_reset,              /* event launch if basis reset            */
    primme_event_converged,          /* report new pair marked as converged    */
-   primme_event_locked              /* report new pair marked as locked       */
+   primme_event_locked,             /* report new pair marked as locked       */
+   primme_event_message,            /* report warning                         */
+   primme_event_profile             /* report time from consumed by a function*/          
 } primme_event;
 
 /* Orthogonalization variant */
@@ -100,24 +95,43 @@ typedef enum {
    primme_orth_explicit_I           /* explicitly compute V'*B*V */
 } primme_orth;
 
+/* Datatype of vectors passed on matrixMatvec, applyPreconditioner,           */
+/* massMatrixMatvec and globalSumReal                                         */
+typedef enum {
+   primme_op_default,   /* The same type as the primme call */
+   primme_op_half,
+   primme_op_float,
+   primme_op_double,
+   primme_op_quad,
+   primme_op_int
+} primme_op_datatype;
+
 typedef struct primme_stats {
    PRIMME_INT numOuterIterations;
    PRIMME_INT numRestarts;
    PRIMME_INT numMatvecs;
    PRIMME_INT numPreconds;
    PRIMME_INT numGlobalSum;         /* times called globalSumReal */
+   PRIMME_INT numBroadcast;         /* times called broadcastReal */
    PRIMME_INT volumeGlobalSum;      /* number of SCALARs reduced by globalSumReal */
+   PRIMME_INT volumeBroadcast;      /* number of SCALARs broadcast by broadcastReal */
+   double flopsDense;               /* FLOPS done by Num_update_VWXR_Sprimme */
    double numOrthoInnerProds;       /* number of inner prods done by Ortho */
    double elapsedTime; 
    double timeMatvec;               /* time expend by matrixMatvec */
    double timePrecond;              /* time expend by applyPreconditioner */
    double timeOrtho;                /* time expend by ortho  */
    double timeGlobalSum;            /* time expend by globalSumReal  */
+   double timeBroadcast;            /* time expend by broadcastReal  */
+   double timeDense;                /* time expend by Num_update_VWXR_Sprimme */
    double estimateMinEVal;          /* the leftmost Ritz value seen */
    double estimateMaxEVal;          /* the rightmost Ritz value seen */
    double estimateLargestSVal;      /* absolute value of the farthest to zero Ritz value seen */
+   double estimateBNorm;            /* estimation of norm of B */
+   double estimateInvBNorm;         /* estimation of norm of inv(B) */
    double maxConvTol;               /* largest norm residual of a locked eigenpair */
    double estimateResidualError;    /* accumulated error in V and W */
+   PRIMME_INT lockingIssue;         /* Some converged with a weak criterion */
 } primme_stats;
 
 typedef struct JD_projectors {
@@ -144,7 +158,6 @@ typedef struct correction_params {
 
 
 typedef struct restarting_params {
-   primme_restartscheme scheme;
    int maxPrevRetain;
 } restarting_params;
 
@@ -157,16 +170,19 @@ typedef struct primme_params {
    void (*matrixMatvec)
       ( void *x, PRIMME_INT *ldx, void *y, PRIMME_INT *ldy, int *blockSize,
         struct primme_params *primme, int *ierr);
+   primme_op_datatype matrixMatvec_type; /* expected type of x and y */
 
    /* Preconditioner applied on block of vectors (if available) */
    void (*applyPreconditioner)
       ( void *x, PRIMME_INT *ldx,  void *y, PRIMME_INT *ldy, int *blockSize,
         struct primme_params *primme, int *ierr);
+   primme_op_datatype applyPreconditioner_type; /* expected type of x and y */
 
    /* Matrix times a multivector for mass matrix B for generalized Ax = xBl */
    void (*massMatrixMatvec)
       ( void *x, PRIMME_INT *ldx, void *y, PRIMME_INT *ldy, int *blockSize,
         struct primme_params *primme, int *ierr);
+   primme_op_datatype massMatrixMatvec_type; /* expected type of x and y */
 
    /* input for the following is only required for parallel programs */
    int numProcs;
@@ -176,6 +192,10 @@ typedef struct primme_params {
    void (*globalSumReal)
       (void *sendBuf, void *recvBuf, int *count, struct primme_params *primme,
        int *ierr );
+   primme_op_datatype globalSumReal_type; /* expected type of sendBuf and recvBuf */
+   void (*broadcastReal)(
+         void *buffer, int *count, struct primme_params *primme, int *ierr);
+   primme_op_datatype broadcastReal_type; /* expected type of buffer */
 
    /*Though primme_initialize will assign defaults, most users will set these */
    int numEvals;          
@@ -193,20 +213,20 @@ typedef struct primme_params {
    int maxBlockSize;
    PRIMME_INT maxMatvecs;
    PRIMME_INT maxOuterIterations;
-   int intWorkSize;
-   size_t realWorkSize;
    PRIMME_INT iseed[4];
-   int *intWork;
-   void *realWork;
    double aNorm;
+   double BNorm;                 /* Approximate 2-norm of B */
+   double invBNorm;              /* Approximate 2-norm of inv(B) */
    double eps;
    primme_orth orth;
+   primme_op_datatype internalPrecision; /* force primme to work in that precision */
 
    int printLevel;
    FILE *outputFile;
 
    void *matrix;
    void *preconditioner;
+   void *massMatrix;
    double *ShiftsForPreconditioner;
    primme_init initBasisMode;
    PRIMME_INT ldevecs;
@@ -219,13 +239,17 @@ typedef struct primme_params {
 
    void (*convTestFun)(double *eval, void *evec, double *rNorm, int *isconv, 
          struct primme_params *primme, int *ierr);
+   primme_op_datatype convTestFun_type; /* expected type of evec */
    void *convtest;
    void (*monitorFun)(void *basisEvals, int *basisSize, int *basisFlags,
-      int *iblock, int *blockSize, void *basisNorms, int *numConverged,
-      void *lockedEvals, int *numLocked, int *lockedFlags, void *lockedNorms,
-      int *inner_its, void *LSRes, primme_event *event,
-      struct primme_params *primme, int *err);
+         int *iblock, int *blockSize, void *basisNorms, int *numConverged,
+         void *lockedEvals, int *numLocked, int *lockedFlags, void *lockedNorms,
+         int *inner_its, void *LSRes, const char *msg, double *time,
+         primme_event *event, struct primme_params *primme, int *err);
+   primme_op_datatype monitorFun_type; /* expected type of float-point arrays */
    void *monitor;
+   void *queue;      /* magma device queue (magma_queue_t*) */
+   const char *profile; /* regex expression with functions to monitor times */
 } primme_params;
 /*---------------------------------------------------------------------------*/
 
@@ -248,85 +272,116 @@ typedef enum {
    PRIMME_LOBPCG_OrthoBasis_Window
 } primme_preset_method;
 
+
+/* Indicates the type of primme_param's members. Used in primme_member_info */
+/* and in primme_svds_member_info                                           */
+
 typedef enum {
    primme_int,
    primme_double,
-   primme_pointer
+   primme_pointer,
+   primme_string
 } primme_type;
 
 typedef enum {
-   PRIMME_n =  0,
-   PRIMME_matrixMatvec =  1,
-   PRIMME_applyPreconditioner =  2,
-   PRIMME_numProcs =  3,
-   PRIMME_procID =  4,
-   PRIMME_commInfo =  5,
-   PRIMME_nLocal =  6,
-   PRIMME_globalSumReal =  7,
-   PRIMME_numEvals =  8,
-   PRIMME_target =  9,
-   PRIMME_numTargetShifts =  10,
-   PRIMME_targetShifts =  11,
-   PRIMME_locking =  12,
-   PRIMME_initSize =  13,
-   PRIMME_numOrthoConst =  14,
-   PRIMME_maxBasisSize =  15,
-   PRIMME_minRestartSize =  16,
-   PRIMME_maxBlockSize =  17,
-   PRIMME_maxMatvecs =  18,
-   PRIMME_maxOuterIterations =  19,
-   PRIMME_intWorkSize =  20,
-   PRIMME_realWorkSize =  21,
-   PRIMME_iseed =  22,
-   PRIMME_intWork =  23,
-   PRIMME_realWork =  24,
-   PRIMME_aNorm =  25,
-   PRIMME_eps =  26,
-   PRIMME_printLevel =  27,
-   PRIMME_outputFile =  28,
-   PRIMME_matrix =  29,
-   PRIMME_preconditioner =  30,
-   PRIMME_initBasisMode =   301,
-   PRIMME_projectionParams_projection =  302,
-   PRIMME_restartingParams_scheme =  31,
-   PRIMME_restartingParams_maxPrevRetain =  32,
-   PRIMME_correctionParams_precondition =  33,
-   PRIMME_correctionParams_robustShifts =  34,
-   PRIMME_correctionParams_maxInnerIterations =  35,
-   PRIMME_correctionParams_projectors_LeftQ =  36,
-   PRIMME_correctionParams_projectors_LeftX =  37,
-   PRIMME_correctionParams_projectors_RightQ =  38,
-   PRIMME_correctionParams_projectors_RightX =  39,
-   PRIMME_correctionParams_projectors_SkewQ =  40,
-   PRIMME_correctionParams_projectors_SkewX =  41,
-   PRIMME_correctionParams_convTest =  42,
-   PRIMME_correctionParams_relTolBase =  43,
-   PRIMME_stats_numOuterIterations =  44,
-   PRIMME_stats_numRestarts =  45,
-   PRIMME_stats_numMatvecs =  46,
-   PRIMME_stats_numPreconds =  47,
-   PRIMME_stats_numGlobalSum =  471,
-   PRIMME_stats_volumeGlobalSum =  472,
-   PRIMME_stats_numOrthoInnerProds =  473,
-   PRIMME_stats_elapsedTime =  48,
-   PRIMME_stats_timeMatvec =  4801,
-   PRIMME_stats_timePrecond =  4802,
-   PRIMME_stats_timeOrtho =  4803,
-   PRIMME_stats_timeGlobalSum =  4804,
-   PRIMME_stats_estimateMinEVal =  481,
-   PRIMME_stats_estimateMaxEVal =  482,
-   PRIMME_stats_estimateLargestSVal =  483,
-   PRIMME_stats_maxConvTol =  484,
-   PRIMME_dynamicMethodSwitch = 49,
-   PRIMME_massMatrixMatvec =  50,
-   PRIMME_convTestFun =  51,
-   PRIMME_convtest =  510,
-   PRIMME_ldevecs =  52,
-   PRIMME_ldOPs =  53,
-   PRIMME_monitorFun = 54,
-   PRIMME_monitor = 55
+   /* NOTE: you can maintain the column of numbers with g+Ctrl-A in vim */
+   PRIMME_n                                      = 1  ,
+   PRIMME_matrixMatvec                           = 2  ,
+   PRIMME_matrixMatvec_type                      = 3  ,
+   PRIMME_applyPreconditioner                    = 4  ,
+   PRIMME_applyPreconditioner_type               = 5  ,
+   PRIMME_massMatrixMatvec                       = 6  ,
+   PRIMME_massMatrixMatvec_type                  = 7  ,
+   PRIMME_numProcs                               = 8  ,
+   PRIMME_procID                                 = 9  ,
+   PRIMME_commInfo                               = 10  ,
+   PRIMME_nLocal                                 = 11  ,
+   PRIMME_globalSumReal                          = 12  ,
+   PRIMME_globalSumReal_type                     = 13  ,
+   PRIMME_broadcastReal                          = 14  ,
+   PRIMME_broadcastReal_type                     = 15  ,
+   PRIMME_numEvals                               = 16  ,
+   PRIMME_target                                 = 17  ,
+   PRIMME_numTargetShifts                        = 18  ,
+   PRIMME_targetShifts                           = 19  ,
+   PRIMME_locking                                = 20  ,
+   PRIMME_initSize                               = 21  ,
+   PRIMME_numOrthoConst                          = 22  ,
+   PRIMME_maxBasisSize                           = 23  ,
+   PRIMME_minRestartSize                         = 24  ,
+   PRIMME_maxBlockSize                           = 25  ,
+   PRIMME_maxMatvecs                             = 26  ,
+   PRIMME_maxOuterIterations                     = 27  ,
+   PRIMME_iseed                                  = 28  ,
+   PRIMME_aNorm                                  = 29  ,
+   PRIMME_BNorm                                  = 30  ,
+   PRIMME_invBNorm                               = 31  ,
+   PRIMME_eps                                    = 32  ,
+   PRIMME_orth                                   = 33  ,
+   PRIMME_internalPrecision                      = 34  ,
+   PRIMME_printLevel                             = 35  ,
+   PRIMME_outputFile                             = 36  ,
+   PRIMME_matrix                                 = 37  ,
+   PRIMME_massMatrix                             = 38  ,
+   PRIMME_preconditioner                         = 39  ,
+   PRIMME_ShiftsForPreconditioner                = 40  ,
+   PRIMME_initBasisMode                          = 41  ,
+   PRIMME_projectionParams_projection            = 42  ,
+   PRIMME_restartingParams_maxPrevRetain         = 43  ,
+   PRIMME_correctionParams_precondition          = 44  ,
+   PRIMME_correctionParams_robustShifts          = 45  ,
+   PRIMME_correctionParams_maxInnerIterations    = 46  ,
+   PRIMME_correctionParams_projectors_LeftQ      = 47  ,
+   PRIMME_correctionParams_projectors_LeftX      = 48  ,
+   PRIMME_correctionParams_projectors_RightQ     = 49  ,
+   PRIMME_correctionParams_projectors_RightX     = 50  ,
+   PRIMME_correctionParams_projectors_SkewQ      = 51  ,
+   PRIMME_correctionParams_projectors_SkewX      = 52  ,
+   PRIMME_correctionParams_convTest              = 53  ,
+   PRIMME_correctionParams_relTolBase            = 54  ,
+   PRIMME_stats_numOuterIterations               = 55  ,
+   PRIMME_stats_numRestarts                      = 56  ,
+   PRIMME_stats_numMatvecs                       = 57  ,
+   PRIMME_stats_numPreconds                      = 58  ,
+   PRIMME_stats_numGlobalSum                     = 59  ,
+   PRIMME_stats_volumeGlobalSum                  = 60  ,
+   PRIMME_stats_numBroadcast                     = 61  ,
+   PRIMME_stats_volumeBroadcast                  = 62  ,
+   PRIMME_stats_flopsDense                       = 63  ,
+   PRIMME_stats_numOrthoInnerProds               = 64  ,
+   PRIMME_stats_elapsedTime                      = 65  ,
+   PRIMME_stats_timeMatvec                       = 66  ,
+   PRIMME_stats_timePrecond                      = 67  ,
+   PRIMME_stats_timeOrtho                        = 68  ,
+   PRIMME_stats_timeGlobalSum                    = 69  ,
+   PRIMME_stats_timeBroadcast                    = 70  ,
+   PRIMME_stats_timeDense                        = 71  ,
+   PRIMME_stats_estimateMinEVal                  = 72  ,
+   PRIMME_stats_estimateMaxEVal                  = 73  ,
+   PRIMME_stats_estimateLargestSVal              = 74  ,
+   PRIMME_stats_estimateBNorm                    = 75  ,
+   PRIMME_stats_estimateInvBNorm                 = 76  ,
+   PRIMME_stats_maxConvTol                       = 77  ,
+   PRIMME_stats_lockingIssue                     = 78  ,
+   PRIMME_dynamicMethodSwitch                    = 79  ,
+   PRIMME_convTestFun                            = 80  ,
+   PRIMME_convTestFun_type                       = 81  ,
+   PRIMME_convtest                               = 82  ,
+   PRIMME_ldevecs                                = 83  ,
+   PRIMME_ldOPs                                  = 84  ,
+   PRIMME_monitorFun                             = 85  ,
+   PRIMME_monitorFun_type                        = 86  ,
+   PRIMME_monitor                                = 87  ,
+   PRIMME_queue                                  = 88  ,
+   PRIMME_profile                                = 89  
 } primme_params_label;
 
+/* Hermitian operator */
+
+int hprimme(PRIMME_HALF *evals, PRIMME_HALF *evecs, PRIMME_HALF *resNorms, 
+      primme_params *primme);
+int kprimme(PRIMME_HALF *evals, PRIMME_COMPLEX_HALF *evecs, PRIMME_HALF *resNorms, 
+      primme_params *primme);
 int sprimme(float *evals, float *evecs, float *resNorms, 
       primme_params *primme);
 int cprimme(float *evals, PRIMME_COMPLEX_FLOAT *evecs, float *resNorms, 
@@ -335,6 +390,50 @@ int dprimme(double *evals, double *evecs, double *resNorms,
       primme_params *primme);
 int zprimme(double *evals, PRIMME_COMPLEX_DOUBLE *evecs, double *resNorms, 
       primme_params *primme);
+int magma_hprimme(PRIMME_HALF *evals, PRIMME_HALF *evecs, PRIMME_HALF *resNorms, 
+      primme_params *primme);
+int magma_kprimme(PRIMME_HALF *evals, PRIMME_COMPLEX_HALF *evecs, PRIMME_HALF *resNorms, 
+      primme_params *primme);
+int magma_sprimme(float *evals, float *evecs, float *resNorms, 
+      primme_params *primme);
+int magma_cprimme(float *evals, PRIMME_COMPLEX_FLOAT *evecs, float *resNorms, 
+      primme_params *primme);
+int magma_dprimme(double *evals, double *evecs, double *resNorms, 
+      primme_params *primme);
+int magma_zprimme(double *evals, PRIMME_COMPLEX_DOUBLE *evecs, double *resNorms, 
+      primme_params *primme);
+
+int hsprimme(float *evals, PRIMME_HALF *evecs, float *resNorms, 
+      primme_params *primme);
+int ksprimme(float *evals, PRIMME_COMPLEX_HALF *evecs, float *resNorms, 
+      primme_params *primme);
+int magma_hsprimme(float *evals, PRIMME_HALF *evecs, float *resNorms, 
+      primme_params *primme);
+int magma_ksprimme(float *evals, PRIMME_COMPLEX_HALF *evecs, float *resNorms, 
+      primme_params *primme);
+
+/* Normal operator */
+
+int kprimme_normal(PRIMME_COMPLEX_HALF *evals, PRIMME_COMPLEX_HALF *evecs, PRIMME_HALF *resNorms, 
+      primme_params *primme);
+int cprimme_normal(PRIMME_COMPLEX_FLOAT *evals, PRIMME_COMPLEX_FLOAT *evecs, float *resNorms, 
+      primme_params *primme);
+int zprimme_normal(PRIMME_COMPLEX_DOUBLE *evals, PRIMME_COMPLEX_DOUBLE *evecs, double *resNorms, 
+      primme_params *primme);
+int magma_kprimme_normal(PRIMME_COMPLEX_HALF *evals, PRIMME_COMPLEX_HALF *evecs, PRIMME_HALF *resNorms, 
+      primme_params *primme);
+int magma_cprimme_normal(PRIMME_COMPLEX_FLOAT *evals, PRIMME_COMPLEX_FLOAT *evecs, float *resNorms, 
+      primme_params *primme);
+int magma_zprimme_normal(PRIMME_COMPLEX_DOUBLE *evals, PRIMME_COMPLEX_DOUBLE *evecs, double *resNorms, 
+      primme_params *primme);
+
+int kcprimme_normal(PRIMME_COMPLEX_FLOAT *evals, PRIMME_COMPLEX_HALF *evecs, float *resNorms, 
+      primme_params *primme);
+int magma_kcprimme_normal(PRIMME_COMPLEX_FLOAT *evals, PRIMME_COMPLEX_HALF *evecs, float *resNorms, 
+      primme_params *primme);
+
+primme_params* primme_params_create(void);
+int primme_params_destroy(primme_params *primme);
 void primme_initialize(primme_params *primme);
 int  primme_set_method(primme_preset_method method, primme_params *params);
 void primme_display_params(primme_params primme);
